
from functools import reduce
from .tree import Node, get_annotation
from funcy import omit
from .constants import *


TO_SKIP = [ELLIPSIS, ADDITIONAL_KEYS]

def _make_schema(node, definitions, root):

    if not len(node.children):
        raise Exception(f'missing definition {repr(node.value)} {"after " + repr(node.parent.value) if node.parent else None}')

    elif any([node.children[0].value == x for x in definitions]): # custom definition
        return {
            '$ref': f'#/definitions/{node.children[0].value}',
            'title': get_title(node),
            'description': get_annotation(node),
        }
    elif node.children[0].value == ELLIPSIS:
        return { 
            'type': 'object', 
            'additionalProperties': True, 
            'title': get_title(node), 
            'description': get_annotation(node),
        }

    elif '..' in node.children[0].value:
        return make_range_schema(node)

    elif '"' in node.children[0].value:
        value = node.children[0].value.split('"')
        value = [x.strip() for x in value if x.strip()] or ['']
        value = value[0]
        return {
            'enum': [value],
            'type': 'string',
            'title': get_title(node),
            'description': get_annotation(node),
        }

    elif node.children[0].value == OR:
        if all(['"' in node.value for node in node.children[0].children]):
            return {
                'enum': [c.value.replace('"', '', 2) for c in node.children[0].children],
                'type': 'string',
                'title': get_title(node),
                'description': get_annotation(node),
            }
        elif all([node.value.isdigit() for node in node.children[0].children]):
            return {
                'enum': [int(c.value) for c in node.children[0].children],
                'type': 'number',
                'title': get_title(node),
                'multipleOf': 1,
                'description': get_annotation(node),
            }
        else:
            return {
                'anyOf': [_make_schema(Node('').insert(c), definitions, root) for c in node.children[0].children],
                'description': get_annotation(node),
            }

    elif node.children[0].value == AND:
        if (len(node.children) > 1): # inline and, interface
            options = []
            options += [_make_schema(Node('').insert(c), definitions, root) for c in node.children[0].children]
            interface_keys = reduce(lambda a, b: [*a, *get_properties(b, root)], options, [])
            children = [c for c in node.children if c.value != AND and c.value not in interface_keys]
            if children:
                inline_props = Node('',).insert(*children)
                options += [_make_schema(inline_props, definitions, root)]
            return {
                'type': 'object',
                'allOf': options,
                'description': get_annotation(node),
            }
        else:
            options = [_make_schema(Node('').insert(c), definitions, root) for c in node.children[0].children]
            return {
                'allOf': options,
                'description': get_annotation(node),
            }

    elif node.children[0].value == LIST:
        return {
            'type': 'array',
            'title': get_title(node),
            'items': _make_schema(Node(node.value, node.parent,).append(node.children[0].children), definitions, root),
            'minItems': 1 if node.children[0].not_empty else 0,
            # 'description': get_annotation(node),
        }

    elif node.children[0].value == STR or node.children[0].value == STRING:
        
        obj = { 
            'type': 'string', 
            'title': get_title(node), 
            'description': get_annotation(node),
        }
        format = get_format(node.value)
        if format:
            obj.update({
                'format': format,
            })
        if node.children[0].not_empty:
            obj.update({
                'minLength': 1,
            })
        return obj

    elif node.children[0].value == REGEX:
        return { 
            'type': 'string', 
            'pattern': node.children[0].pattern, 
            'title': get_title(node), 
            'description': get_annotation(node),
        }

    elif node.children[0].value == ANY:
        return { 
            'title': get_title(node),
            'description': get_annotation(node), 
        }

    elif node.children[0].value == BOOL:
        return { 
            'type': 'boolean', 
            'title': get_title(node),
            'description': get_annotation(node),
        }

    elif node.children[0].value == NULL:
        return { 'const': None }

    elif node.children[0].value == FLOAT:
        return { 
            'type': 'number', 
            'title': get_title(node), 
            'description': get_annotation(node),
        }

    elif node.children[0].value == INT:
        return { 
            'type': 'number',
            "multipleOf": 1.0,
            'title': get_title(node),
            'description': get_annotation(node),
        }
    else: # object
        ellipses = [ x for x in node.children if x.value == ELLIPSIS ]
        additional_props = next((_make_schema(child, definitions, root) for child in node.children if child.value == ADDITIONAL_KEYS), None)
        obj = {
            'title': get_title(node),
            'description': get_annotation(node),
            'type': 'object',
            'required': [child.value for child in node.children if child.required and not child.value in TO_SKIP],
            'properties': {
                child.value: _make_schema(child, definitions, root) for child in node.children if not child.value in TO_SKIP
            },
            'additionalProperties': omit(additional_props, ['title']) if additional_props != None else bool(ellipses)
        }
        # if not ellipses or additional_props:
        #     obj = omit(obj, ['additionalProperties'])
        return obj

def get_title(node):
    return node.value if node.value not in [LIST, OR, AND, REGEX,] else ''


# TODO
def make_schema(root):
    definitions = [child.value for child in root.children]
    # print(definitions)
    schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        '$ref': '#/definitions/' + definitions[0],
        'definitions': {},
    }
    for child in root.children:
        schema['definitions'][child.value] = _make_schema(child, definitions, root)

    return schema


def get_format(name):
    if name == 'DateTime':
        return 'date-time'
    if name == 'Date':
        return 'date'
    if name == 'Time':
        return 'time'
    if name == 'Email':
        return 'email'
    if name == 'Uri':
        return 'uri'
    if name == 'Iri':
        return 'iri'
    return None

def make_range_schema(node):
    value = node.children[0].value
    boundaries = value.split('..')
    boundaries = [b.strip() for b in boundaries]
    boundaries = [b for b in boundaries if b]
    obj = {
        'title': get_title(node),
        'type': 'number',
        'description': get_annotation(node),
    }
    if all([s.isdigit() for s in boundaries]):
        obj.update({
            'multipleOf': 1,
        })
        boundaries = [int(b) for b in boundaries]
    else:
        boundaries = [float(b) for b in boundaries]
    if len(boundaries) == 0:
        raise Exception('range must contain at least one bounding: 0.. ..100 0..100')
    elif len(boundaries) <= 1:
        if value.index('..') == 0:
            obj.update({
                'type': 'number',
                'exclusiveMaximum': boundaries[0],
            })
        else:
            obj.update({
                'type': 'number',
                'minimum': boundaries[0],
            })
    else:
        obj.update({
            'type': 'number',
            'minimum': boundaries[0],
            'exclusiveMaximum': boundaries[1],
        })
    return obj


def get_properties(block: dict, root: Node):
    if block.get('properties', {}):
        return block['properties']
    elif block.get('$ref', ''):
        ref_name = block.get('$ref', '').split('/')[-1]
        node = [n for n in root.children if n.value == ref_name][0]
        # print(node.children)
        return [c.value for c in node.children if c.value not in [OR, AND, LIST,]]
    else:
        raise Exception(f'cannot get properties from {block}')


import json
from funcy import distinct
from dataclasses import dataclass, fields
from .tree import Node
from .constants import *
from .support import capitalize
from typing import Union, List
from .resolve_refs import resolve_refs

@dataclass
class SchemaBlock:
    previous_key: str = ''
    data: str = ''
    type: Union[str, List[str]] = None
    items: list = None
    properties: dict = None 
    required: list = None
    anyOf: list = None
    enum: list = None
    allOf: list = None
    oneOf: list = None
    format: str = None
    @classmethod
    def make(cls, data, previous_key=''):            
        p = [f.name for f in fields(SchemaBlock)]
        x = {k:v for k, v in data.items() if k in p}
        return cls(**x, data=json.dumps(data, indent=4), previous_key=previous_key)

def schema_to_tree(schema: SchemaBlock, node, references=[]):
    """
    type == integer -> node.insert(Node(INT, node))
    type == number -> node.insert(Node(FLOAT, node))
    type == string -> node.insert(Node(STR, node))
    type == array -> Node(LIST, node).insert(schema_to_tree(item, node))
    type == object -> Node(LIST, node).insert(schema_to_tree(item, node))
    type == or -> [OrNode.insert(schema_to_tree(item)) for item in items]
    """
    if False:
        pass
    elif schema.anyOf or schema.oneOf or schema.allOf or schema.enum:
        child = Node(AND if schema.allOf else OR, node)
        items = schema.oneOf or schema.anyOf or schema.allOf or schema.enum
        items = distinct(items, key=get_type_name)
        i = 0
        for subset in items:
            if isinstance(subset, str):
                subset = SchemaBlock(type=subset,)
            else:
                subset = SchemaBlock.make(subset,)
            if subset.type in ['string', 'integer', 'number', 'boolean', 'null']:
                child = schema_to_tree(subset, child, references)
            else:
                if all([node.value != x for x in (LIST, OR, AND)]):
                    reference_name = capitalize(node.value) + str(i)
                else:
                    reference_name = capitalize(node.parent.value) + 'Elem' + str(i)
                reference_root = Node(reference_name, child)
                reference_root = schema_to_tree(subset, reference_root, references)
                references.append(reference_root)
                child = child.insert(Node(reference_name, child))
                i += 1
        node = node.insert(child,)
    elif schema.type == 'object':
        properties = schema.properties or {}
        if properties:
            for k, v in properties.items() or []:
                v = SchemaBlock.make(v, previous_key=k)
                required = k in (schema.required or [])
                child = Node(k, node, required=required)
                child = schema_to_tree(v, child, references)
                node = node.insert(child)
        else:
            child = Node(ANY, node)
            node = node.insert(child)
        
    elif schema.type == 'array':
        subset = v = SchemaBlock.make(schema.items or {})
        child = Node(LIST, node)
        child = schema_to_tree(subset, child, references)
        node = node.insert(child)
        
    elif schema.type == 'string':
        child = Node(STR, node)
        node = node.insert(child)
    elif schema.type == 'integer':
        child = Node(INT, node)
        node = node.insert(child)
    elif schema.type == 'number':
        child = Node(FLOAT, node)
        node = node.insert(child)
    elif schema.type == 'boolean':
        child = Node(BOOL, node)
        node = node.insert(child)
    
    elif isinstance(schema.type, list):
        child = Node(OR, node)
        items = distinct(schema.type, key=get_type_name)
        for subset in items:
            if 'type' in subset:
                subset = SchemaBlock.make(subset)
                child = schema_to_tree(subset, child, references)
            elif isinstance(subset, str):
                subset = SchemaBlock(type=subset)
                child = schema_to_tree(subset, child, references)
            else:
                raise NotImplementedError(str(subset))
        node = node.insert(child,)
    else:
        # raise NotImplementedError(str(schema))
        # print('!!!NotImplementedError!!!', schema.previous_key)
        # print(str(schema))
        child = Node(ANY, node)
        node = node.insert(child)
    return node


def get_type_name(schema):
    if isinstance(schema, str):
        return 'string'
    if isinstance(schema, dict):
        if 'type' in schema:
            if schema['type'] == 'object':
                return str(schema)
            return schema['type']
    if any([x in schema for x in ('anyOf', 'allOf', 'enum', 'oneOf',)]):
        return str(schema)
    return 'any'



def from_jsonschema_to_tree(schema, ref_name=None, ):
    references: list = []
    schema = {**schema}
    resolve_refs(schema, add_titles=True)
    # print(json.dumps(schema, indent=4))
    root_name = schema.get('title') or ref_name or 'Root'
    root_name = root_name.replace('_', ' ')
    root_name = ''.join([capitalize(x) for x in root_name.split(' ')])
    schema = SchemaBlock.make(schema)
    t = schema_to_tree(schema, node=Node(root_name), references=references)
    return t, references

def from_jsonschema(schema, ref_name=None):
    t, references = from_jsonschema_to_tree(schema, ref_name)
    res = ''
    res += t.to_skema()
    for ref in references:
        res += '\n\n'
        res += ref.to_skema()
    res += '\n'
    return res


from .split_tree_parts import split_tree_parts
from .tree import Node


scalar_already_present = [ # empty to make tests pass
    'ID',
    # 'Json', 
    # 'DateTime',
    # 'Time',
    # 'Date',
]

def to_graphql(string: str, hide=scalar_already_present, only=None) -> str:
    refs = split_tree_parts(string, language='graphql', hide=hide, only=only,)
    if not 'Json' in hide and not (only and not 'Jons' in only):
        refs += [Node('Json',).append([Node('')])] 
    types = [t.to_graphql() for t in refs]
    schema = '\n\n'.join(types)
    return schema


# printf "\nA:\n  b: Str\n" | xargs -0I%  python -m src %
import sys
import json
from functools import reduce, partial
from .to_jsonschema import to_jsonschema


def main(schema, ref=None, resolve=False):
    return json.dumps(to_jsonschema(schema, ref, resolve), indent=4)

if __name__ == '__main__':
    string = sys.stdin.read()
    print(main(string))






from .tokenize import tokenize
from .make_schema import make_schema
from .make_tree import make_tree
from .resolve_refs import resolve_refs
from .support import rcompose
from copy import deepcopy

def resolve_schema(jsonschema, ref=None):
    jsonschema = deepcopy(jsonschema)
    jsonschema["$ref"] = jsonschema["$ref"] if not ref else "#/definitions/" + ref
    resolve_refs(jsonschema)
    return jsonschema


def to_jsonschema(schema, ref=None, resolve=False, hide=[], only=None):

    node = make_tree(tokenize(schema))
    result = make_schema(node)
    if ref and not ref in result.get("definitions", {}):
        raise Exception(f"can't find definition {ref}")

    if only and isinstance(only, list):
        result["definitions"] = {
            k: v for k, v in result["definitions"].items() if k in only
        }
    if hide and isinstance(hide, list):
        result["definitions"] = {
            k: v for k, v in result["definitions"].items() if not k in hide
        }

    result["$ref"] = result["$ref"] if not ref else "#/definitions/" + ref

    if resolve:
        result = resolve_schema(result, ref)

    return result
